/*===================================================================
  Xwindow Realtime Simulator
  by kitsune(originated by sasa (originated by  gutchi (originated in WTNB)))

===========================================================================*/

//*********** include **************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <iostream.h>
#include <fstream.h>


//********** define *************

#define pi 3.14159265358979
#define PX 2
#define MAX 302
#define MAY 302
#define i_0 1
#define PARAMAX      10
#define DISPMAX      10
#define MCOLOR 65535
#define PHA_C 1                   /* phase no color partition */
#define AMP_C 12                  /* amplitude no color partition */
#define PHI_R pi*2.0/3.0          /* RED phase */
#define PHI_G 0                   /* GREEN phase */
#define PHI_B pi*4.0/3.0          /* BLUE phase */
#define PHI_0 (-pi*2.0/5.0)       /* OFFSET phase */
#define KAPPA 3                   /* AMPLITUDE DUMPING FACTOR */
#define AMPF 0.8                  /* AMPLITUDE FACTOR (int)(|W|^2*AMP_C*AMPF)*/
#define SWINS 100                 /* SAMPLE WINDOW NO OOKISA */
#define SS 1                      /* STOP SIGNAL */
#define DPARAM 1                  /* PARAMETER CHANGE */
#define FUNC (8+2*paranumber)     /* NUMBER OF FUNCTIONS */
#define FX 50                     /* width of each function region */
#define FY 20                     /* height of each function region */


//************ function & class declare ************

void      read_parameter(char *filename);
void      initialize_function(void);
void      win_draw(void);
void      snapshot_save(char*);
void      snapshot_load(char*);

class     XRS_para;
class     XRS_disp;


//************ main class **************

class XRS{

      int     i_max,j_max;
      int     digA;
      int     *changedisp;
      int     paranumber,dispnumber;
      int     id;
      int     *ts; 

      double  *t;
      double  *p[PARAMAX];

      unsigned long  map_old[MAX][MAY],map[MAX][MAY];
      unsigned long  color[PHA_C*AMP_C+2]; 
      unsigned long  black,white;

      char    wanya[20];
      char    paraname[PARAMAX][8],dispname[DISPMAX][8];
 
      Window  root,owin,win,swin,cwin;
      Display *d;
      GC      gc,bgc,wgc,dgc;
      Font    dfont;
      XSetWindowAttributes  att;
      XEvent    event;

  public:
      XRS(){d = XOpenDisplay(NULL);}      
      ~XRS(){XCloseDisplay(d);}
      void  initialize(XRS_para P,XRS_disp D,double &time,
                                    int i,int j,int& disp=0);
      void  initialize(XRS_para P,XRS_disp D,int &step,
                                    int i,int j,int& disp=0);
      void  window_set();
      void  checkevent();
      int   xcommand();
      void  click_location();
      void  initialize_map();
      void  initialize_color();
      void  owin_draw();
      void  sample_draw();
      void  color_draw(double,int,int);
      void  point_draw(int,int,int);
      void  point_dot(int,int,int);
      void  point_erase(int,int);
      void  current_parameter(void);
};


//****** parameter pointer & name set *********

class XRS_para
{
      int i;
      int paranumber;
      double *p[PARAMAX];
      char paraname[PARAMAX][8];        
   public:
      XRS_para(){
         paranumber=0;
         i=0;
      }
      XRS_para& operator<<(double& s){
         p[paranumber]=&s;
         ++paranumber;
         return *this;
      }
      XRS_para& operator<<(char name[8]){
         for(int k=0;k<8;++k) paraname[i][k]=name[k];
         ++i;
         return *this;
      }
      friend void XRS::initialize(XRS_para,XRS_disp,double &,int,int,int&);
      friend void XRS::initialize(XRS_para,XRS_disp,int &,int,int,int&);
};


//******* display name set ***********

class XRS_disp
{
      int dispnumber;
      char dispname[DISPMAX][8];

   public:
      XRS_disp(){dispnumber=0;}
      XRS_disp& operator<<(char name[8]){
         for(int k=0;k<8;++k) dispname[dispnumber][k]=name[k];
         ++dispnumber;
         return *this;
      }
      friend void XRS::initialize(XRS_para,XRS_disp,double &,int,int,int&);
      friend void XRS::initialize(XRS_para,XRS_disp,int &,int,int,int&);
};


//******* initialize *************

void  XRS::initialize(XRS_para P,XRS_disp D,double &time,
                                        int i=100,int j=100,int& disp=0)
{
//------- t,i_max,j_max,change_disp pointer set ---------

      t=&time;
      id=0;
      i_max=i;
      j_max=j;
      changedisp=&disp;
      paranumber=P.paranumber;
      dispnumber=D.dispnumber;
      p=P.p;
      for( int m = 0 ;m < paranumber ;++m)
        for(int k=0;k<8;++k) paraname[m][k]=P.paraname[m][k];
      for( m = 0 ;m < dispnumber ;++m)
        for(int k=0;k<8;++k) dispname[m][k]=D.dispname[m][k];

//------- window setting & initialize -------
 
       window_set();
}


void  XRS::initialize(XRS_para P,XRS_disp D,int &step,
                                        int i=100,int j=100,int& disp=0)
{
//------- ts,i_max,j_max,change_disp pointer set ---------

      ts=&step;
      id=1;
      i_max=i;
      j_max=j;
      changedisp=&disp;
      paranumber=P.paranumber;
      dispnumber=D.dispnumber;
      p=P.p;
      for( int m = 0 ;m < paranumber ;++m)
        for(int k=0;k<8;++k) paraname[m][k]=P.paraname[m][k];
      for( m = 0 ;m < dispnumber ;++m)
        for(int k=0;k<8;++k) dispname[m][k]=D.dispname[m][k];

//------- window setting & initialize -------

       window_set();
}



//******* window_setting & initialize **********

void XRS::window_set(void)
{
//------- window set -------

      root = RootWindow(d, 0);
      dfont = XLoadFont(d, "6x13");

      black = BlackPixel(d, 0); 
      white = WhitePixel(d, 0);

      gc = XCreateGC(d, root, 0, 0);
      bgc = XCreateGC(d, root, 0, 0);
      wgc = XCreateGC(d, root, 0, 0);
      dgc = XCreateGC(d, root, 0, 0);

      XSetLineAttributes(d, gc, 1, LineSolid, CapButt, JoinMiter);
      XSetLineAttributes(d, bgc, 0, LineSolid, CapButt, JoinMiter);
      XSetLineAttributes(d, wgc, 0, LineSolid, CapButt, JoinMiter);
      XSetLineAttributes(d, dgc, 0, LineOnOffDash, CapButt, JoinMiter);

      XSetForeground(d, bgc, black);
      XSetBackground(d, bgc, white);
      XSetForeground(d, wgc, white);
      XSetForeground(d, dgc, black);

      XSetFont(d, bgc, dfont);
      XSetFont(d, wgc, dfont);
      XSetFont(d, dgc, dfont);

      win = XCreateSimpleWindow(d, root, 300, 300,i_max*PX+PX
                               ,j_max*PX+PX, 2, black, white);
      XStoreName(d, win, "Spatial Pattern");

      XMapRaised(d, win);
/*
      cwin = XCreateSimpleWindow(d, root, 120+(i_max+1)*PX, 100, SWINS,
                               SWINS, 2, black, white);
      XStoreName(d, cwin, "Complex Plane");
      XMapRaised(d,cwin);
*/
      swin = XCreateSimpleWindow(d, root, 500, 300, SWINS+1,
                               SWINS+1, 2, black, white);
      XStoreName(d, swin, "Sample Window");
      XMapRaised(d,swin); 

      owin = XCreateSimpleWindow(d, root, 100, 50, FX*((FUNC+1)/2)+1,
                       FY*2+1,2, black, white);
      XStoreName(d, owin, "Growing Pattern");
      XSelectInput(d, owin, ButtonPressMask|StructureNotifyMask);
      XSelectInput(d, win , ButtonPressMask);
      XMapRaised(d, owin);

//----- initialization -----

      initialize_color();
      initialize_map();

      while(event.type != MapNotify)
          XWindowEvent(d, owin, StructureNotifyMask, &event);

      owin_draw();
      current_parameter();
      XDrawImageString(d, owin, bgc,    0+5, FY-4, "       ", 7);
      XDrawImageString(d, owin, bgc,    0+5, FY-4, " START ", 7);

      win_draw();
      sample_draw();

      XFlush(d);
   
      XNextEvent(d, &event);
      while(event.type != ButtonPress)
          XNextEvent(d, &event);
      XDrawImageString(d, owin, bgc,    0+5, FY-4, "       ", 7);
      XDrawImageString(d, owin, bgc,    0+5, FY-4, " QUIT  ", 7);

}


//********* event ***********

inline void  XRS::checkevent(void)
{
      if(XCheckWindowEvent(d, owin, ButtonPress, &event))
          if(xcommand() == SS) exit(1); 

      if(XCheckWindowEvent(d, win, ButtonPress, &event))
          click_location();
      
      XDrawImageString(d, owin, bgc, FX+5, 2*FY-4, "       ", 7);
      
      if(id==0)
          sprintf(wanya, "%f7.2",*t);
      else
          sprintf(wanya, "%7d",*ts);     

      XDrawImageString(d, owin, bgc, FX+5, 2*FY-4, wanya, 7);
}


//********* xcommand ********* 

int  XRS::xcommand(void)
{
      int i,j;
      int fun,xfun,yfun;
      char filename[126];

      xfun=event.xbutton.x/FX;
      yfun=event.xbutton.y/FY;
      fun=2*xfun+yfun;
							
      if (fun == 0){
          /* Quit */
          return(SS);
      
      } else if (fun == 1){
          /* Pause/Cont */
          owin_draw();
          sample_draw();
          XDrawImageString(d, owin, bgc, 5, 2*FY-4, "       ", 7);
          XDrawImageString(d, owin, bgc,   5,   2*FY-4, " CONT  ", 7);
          XWindowEvent(d, owin, ButtonPress,  &event);
          XDrawImageString(d, owin, bgc, 5, 2*FY-4, "       ", 7);
          XDrawImageString(d, owin, bgc,   5,   2*FY-4, " PAUSE ", 7);

      } else if (fun == 2){
          /* Save */
          cout<<"Enter save filename  ";
          cin>>filename;
          snapshot_save(filename);
          cout<<"snaphot.data save >> "<<filename<<"\n";

      } else if (fun == 3){
          /* Time or Step value */
          if(id==0)
             *t=0;
          else
             *ts=0;

      } else if (fun == 4){
          /* Load */
          cout<<"Enter load filename  ";
          cin>>filename;
          snapshot_load(filename);
          cout<<filename<<" >> snapshot_data load"<<"\n";

      } else if (fun == 5){
          /* Read parameters */
          cout<<"Enter parameter filename :";
          cin>>filename;
          read_parameter(filename);
          current_parameter();
          cout<<filename<<" >> parameter_data read"<<"\n";

      } else if (fun == 6){
          /* display */
          XDrawImageString(d, owin, bgc, 3*FX+5,   FY-4, "       ", 7);
          if (*changedisp==(dispnumber-1)) 
                *changedisp=0; 
          else 
                ++(*changedisp);
          XDrawImageString(d,owin, bgc, 3*FX+5, FY-4,dispname[*changedisp] ,7);

      } else if (fun == 7){
          /* initial */
          initialize_function();
          if(id==0)
              *t=0.0;
          else
              *ts=0;

      } else if (fun >= 8 && fun < FUNC){
          /* parameter change */
          
          if (yfun == 0){
              /* parameter up */
              if (event.xbutton.button==1)
                  *p[xfun-4]+=DPARAM;
              else if (event.xbutton.button==2)
                  *p[xfun-4]+=DPARAM*0.1;
              else if (event.xbutton.button==3)
                  *p[xfun-4]+=DPARAM*0.01;
              current_parameter();
 
          }else{
              /* parameter down */
              if (event.xbutton.button==1)
                  *p[xfun-4]-=DPARAM;
              else if (event.xbutton.button==2)
                  *p[xfun-4]-=DPARAM*0.1;
              else if (event.xbutton.button==3)
                  *p[xfun-4]-=DPARAM*0.01;
              current_parameter();
 
          }
     
      } else {
          return(SS);
      }
          
      return(SS-1);

}

//********* click locat*************

inline void  XRS::click_location(void)
{
     /* display button click location in a graphic window */
    cout<<"Click at (i,j)=(  "
        <<event.xbutton.x/PX<<" , "<<event.xbutton.y/PX;
    
    if(id==0)
        cout<<" )  time = "<<*t<<"\n";
    else
        cout<<" ) time step = "<<*ts<<"\n";
}


//********* initialize map ***********

void  XRS::initialize_map(void)
{
      register int i,j;

      for(j=0; j<MAY; j++){
          for(i=0; i<MAX; i++){

              map_old[i][j] = white;

          }
      }
      return;
}


//********* initialize color *********

void  XRS::initialize_color(void)
{
      int i,j,digphi;
      float odias, dc;
      Colormap cm;
      XColor   cc;

      cm = DefaultColormap(d, 0);

      for(j=0; j<AMP_C; j++){

          odias = 1.0-(float)(j)/AMP_C;

/*        odias = exp(-KAPPA*(float)(j)/AMP_C); */

          for(i=0; i<PHA_C; i++){
              digphi=i;

              dc=cos((digphi+0.5)*2*pi/PHA_C+PHI_R+PHI_0);
              if(dc > 0.5)
                  cc.red = (int)MCOLOR;
              else if (dc < -0.5)
                  cc.red = (int)(MCOLOR*odias);
              else
                  cc.red = (int)(MCOLOR*(odias+(1.0-odias)*(dc+0.5)));

              dc=cos((digphi+0.5)*2*pi/PHA_C+PHI_G+PHI_0);
              if(dc > 0.5)
                  cc.green = (int)MCOLOR;
              else if (dc < -0.5)
                  cc.green = (int)(MCOLOR*odias);
              else
                  cc.green = (int)(MCOLOR*(odias+(1.0-odias)*(dc+0.5)));

              dc=cos((digphi+0.5)*2*pi/PHA_C+PHI_B+PHI_0);
              if(dc > 0.5)
                  cc.blue = (int)MCOLOR;
              else if (dc < -0.5)
                  cc.blue = (int)(MCOLOR*odias);
              else
                  cc.blue = (int)(MCOLOR*(odias+(1.0-odias)*(dc+0.5)));

      /* mono_tone */
/*
              cc.red = (int)(MCOLOR*odias);
              cc.green = (int)(MCOLOR*odias);
              cc.blue = (int)(MCOLOR);
*/

              XAllocColor(d, cm, &cc);
              color[j*PHA_C+i] = cc.pixel;

          }
      }
      cc.red=0;
      cc.green=0;
      cc.blue=0;
      XAllocColor(d, cm, &cc);
      color[AMP_C*PHA_C] = cc.pixel;

      cc.red=MCOLOR;
      cc.green=MCOLOR;
      cc.blue=MCOLOR;
      XAllocColor(d, cm, &cc);
      color[AMP_C*PHA_C+1] = cc.pixel;

      return;
}


//********* owin_draw *********

void  XRS::owin_draw(void)
{
      XDrawLine(d, owin, bgc, 0, FY, FX*(FUNC+1)/2+1, FY);
      for (int i=1;i<FUNC/2;i++){
           XDrawLine(d, owin, bgc, FX*i, 0, FX*i, FY*2);
      }

      XDrawImageString(d, owin, bgc,    0+5,   FY-4, " QUIT  ", 7);
      XDrawImageString(d, owin, bgc,    0+5, 2*FY-4, " PAUSE ", 7);
      XDrawImageString(d, owin, bgc,   FX+5,   FY-4, " SAVE  ", 7);

      XDrawImageString(d, owin, bgc, 2*FX+5,   FY-4, " LOAD  ", 7);
      XDrawImageString(d, owin, bgc, 2*FX+5, 2*FY-4, " READ  ", 7);
      XDrawImageString(d, owin, bgc, 3*FX+5,   FY-4,dispname[*changedisp], 7);
      XDrawImageString(d, owin, bgc, 3*FX+5, 2*FY-4, "initial", 7);

      for(i = 0; i<paranumber; ++i){
          int k=i+4;
          XDrawImageString(d, owin, bgc, k*FX+5,   FY-4,  paraname[i], 7);
          int rec=sprintf(wanya, "%f7.2",*p[i]);
          XDrawImageString(d, owin, bgc, k*FX+5, 2*FY-4,  wanya, 7);
      }

      return;
}


//************ sample_draw ***********

void  XRS::sample_draw(void)
{
      double sx,sy;
      double w02,phi0;
      double mono,sashi,m,s;
      mono=AMPF*SWINS/2;
      sashi=mono/5;

      for(int j=i_0; j<SWINS; j++){
          for(int i=i_0; i<SWINS; i++){

              int k=i-SWINS/2; 
              m=k-floor(k/mono)*mono;
              s=k-floor(k/sashi)*sashi;
              sx=(double)(i-SWINS/2)/SWINS*2.0/AMPF;
              sy=(double)(j-SWINS/2)/SWINS*2.0/AMPF;

              digA=(int)((sqrt(sx*sx))*AMP_C*AMPF);
              if (sx*sy==0)                                   /* axis */
                  map[i][j] = color[AMP_C*PHA_C];
              else if ( (int)m==0 && j < 10 )                 /*memori*/
                  map[i][j]=color[AMP_C*PHA_C];
              else if ( (int)s==0 && j < 5 )
                  map[i][j]=color[AMP_C*PHA_C];
              else                                            /* futsu */
                  map[i][j] = color[digA*PHA_C];

              XSetForeground(d, gc, map[i][j]);

              XDrawPoint(d, swin, gc, i, j);
          }
      }

      XSetForeground(d, gc, white);

      sx=sqrt(w02)*cos(phi0);
      sy=sqrt(w02)*sin(phi0);
    
      XDrawPoint(d, swin, gc, (int)(sx/2*AMPF*SWINS+SWINS/2),
               (int)(sy/2*AMPF*SWINS+SWINS/2));
      XDrawPoint(d, swin, gc, (int)(sx/2*AMPF*SWINS+SWINS/2)+1,
               (int)(sy/2*AMPF*SWINS+SWINS/2));
      XDrawPoint(d, swin, gc, (int)(sx/2*AMPF*SWINS+SWINS/2),
               (int)(sy/2*AMPF*SWINS+SWINS/2)+1);
      XDrawPoint(d, swin, gc, (int)(sx/2*AMPF*SWINS+SWINS/2)+1,
               (int)(sy/2*AMPF*SWINS+SWINS/2)+1);

      XDrawPoint(d, swin, gc, (int)(-sx/2*AMPF*SWINS+SWINS/2),
               (int)(-sy/2*AMPF*SWINS+SWINS/2));
      XDrawPoint(d, swin, gc, (int)(-sx/2*AMPF*SWINS+SWINS/2)-1,
               (int)(-sy/2*AMPF*SWINS+SWINS/2));
      XDrawPoint(d, swin, gc, (int)(-sx/2*AMPF*SWINS+SWINS/2),
               (int)(-sy/2*AMPF*SWINS+SWINS/2)-1);
      XDrawPoint(d, swin, gc, (int)(-sx/2*AMPF*SWINS+SWINS/2)-1,
               (int)(-sy/2*AMPF*SWINS+SWINS/2)-1);

      return;
}


//************* color_draw ***********

inline void  XRS::color_draw(double W,int i,int j)
{
      digA=(int)(W*W*AMP_C*AMPF);
      if(digA > AMP_C) digA=(int)(AMP_C);

      map[i][j] = color[digA*PHA_C];

      if(map[i][j] != map_old[i][j]){
            XSetForeground(d, gc, map[i][j]);

            XDrawPoint(d, win, gc, i*PX,   j*PX);
            XDrawPoint(d, win, gc, i*PX+1, j*PX);
            XDrawPoint(d, win, gc, i*PX,   j*PX+1);
            XDrawPoint(d, win, gc, i*PX+1, j*PX+1);

            map_old[i][j] = map[i][j];
      }
}


//************ point_draw (for 1D graphic) ***********

inline void  XRS::point_draw(int i,int j,int k)
{
      if(j != (map_old[i][k]>>1)){
          point_erase(i,k);
          point_dot(i,j,k);
      }
}          

//************ point_dot (for 1D graphic) ***********

inline void  XRS::point_dot(int i,int j,int k)
{
      XSetForeground(d, gc, color[k]);
      XDrawPoint(d, win, gc, i*PX,j_max*PX-j);
        
      map_old[i][k] = (j<<1);
}


//************ point_erase (for 1D graphic) ***********            

inline void  XRS::point_erase(int i,int k)
{
      int j;

      j = (map_old[i][k]>>1);
      XSetForeground(d, gc, white);
      XDrawPoint(d, win, gc, i*PX,j_max*PX-j);
}


//************ current_parameter ***********

void  XRS::current_parameter(void)
{
    for(int i=0; i<paranumber; ++i){
        int k=i+4; 
        XDrawImageString(d, owin, bgc, k*FX+5, 2*FY-4, "       ", 7);
        sprintf(wanya, "%f7.2",*p[i]);
        XDrawImageString(d, owin, bgc, k*FX+5, 2*FY-4,  wanya, 7);
    }
    return;
}


