/*==========================================================================
   Xwindow Realtime Simulator Sample (2D simulation)
   by kitsune(originated by sasa (originated by  gutchi (originated in WTNB)))
==========================================================================*/

/*------------------------ comment ----------------------------------

   function declare
      void initialize_function(void);
      void win_draw(void);
      void snapshot_save(char *filename);
      void snapshot_load(char *filename);
      void read_parameter(char *filename);

      are wrriten in a include file "XRS.h".
      So, you have to define above functions.

   define
      #define pi 3.14159265358979
      #define MAX 202
      #define MAY 102
      #define i_0 1
      #define ( other window set parameters )

      are also wrriten in a include file "XRS.h".
      you may use or change, as you like.

  classes defined in XRS.h
     class XRS;       : main class 
     class XRS_para;  : class for parameter & parameter name input 
     class XRS_disp;  : class for display name input

  main menbers of class XRS (for 2D simulation)
     void initialize(XRS_para,XRS_disp,double& or int&,int=100,int=100,int&=0);
               : initialize windows and 
               : set parameters,dispnames,pointer of t,i_max,j_max,
               :     and pointer of changedisp(defalt 0).
               : t is acceptable to be either type , double or int.  
     void  checkevent();                      
               : check event and display t.
     void  color_draw(double W,int i,int j);
               : draw color at point(i*PX,j*PX) in a graphic window.
               : color density = (int)(W*W*AMP_C*AMPF)

  how to run
     XRSver1sample pqr parameterfile1 [parameterfile2 parameterfile3 ...]
         p=B -> Background
         q=y -> snapshot file save
         r=a -> load snapshot file at first 
   
---------------------------------------------------------------------------*/
  

//*********** include **************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include "/home/home2/kitsune/INCLUDE/XRS.h"


//*********** define ***************

#define V(x)        (alpha*pow((x),beta))
#define RAN()       ((double)rand()/2147483647)
#define RANSU       (RAN()-0.5)
#define integer(x)  ((int)(floor(x)+floor(2*((x)-floor(x)))))


//*********** global variable ************

double                  t;
double                  D[MAX][MAY];
double                  n[MAX][MAY];

double                  LX,LY;
double                  DT,DX;
double                  time_interval;
double                  save_time;
double                  quit_time;

double                  alpha;
double                  beta;
double                  e;
double                  ga;
double                  r_start;
double                  disturb;

char                    name[26]; /* snapshot_save filename */
 
int                     i_max,j_max;
int                     changedisp;   

//********** function & class prototype *****************

void                    n_snapshot_save(int n);
void                    boundary_condition(void);
XRS                     X;
XRS_para                parameter;
XRS_disp                dispname;


//********** main ***********

main(int argc,char *argv[])
{

//----- main routine variable -------
 
     int step,disp_step,save_step,quit_step,paranumber;
     char filename[126];

//----- parameter input -----
    
     for(paranumber=2;paranumber<argc;paranumber++){

        read_parameter(argv[paranumber]);

        i_max = (int)(LX/DX) + i_0;
        j_max = (int)(LY/DX) + i_0;
        if(*argv[1]!='B') time_interval=DT; else time_interval=save_time; 
        disp_step=integer(time_interval/DT);
        save_step=integer(save_time/DT);
        quit_step=integer(quit_time/DT);

//----- initialize ----------

     parameter<<alpha<<beta<<e<<ga;
     parameter<<" alpha "<<"  beta "<<"   e   "<<"   ga  ";
     dispname<<"   D   "<<"   n   ";

     if(*argv[1]!='B')
         X.initialize(parameter,dispname,t,i_max,j_max,changedisp);
     
     t=0;
     initialize_function();

     if(*(argv[1]+2)=='a'){
          cout<<"Enter filename  ";
          cin>>filename;
          snapshot_load(filename);
          cout<<filename<<" >> snapshot_data load"<<"\n";
     }

//----- main routine --------

     while(1){

         int ts;
      
         if(*argv[1]!='B') X.checkevent();

         double   nn[MAX][MAY],D2[MAX][MAY],D3[MAX][MAY];
         double   m,nl,dD,dn,se,sed,tee,txx,txxg;
         se=e/(1.41421356);
         sed=disturb*se;
         tee=DT/(e*e);
         txx=DT/(DX*DX);
         txxg=txx*ga;

 
         for(ts=0; ts<disp_step; ts++){

             register i,j;
 
//----- D local map -----
             for(i=i_0;i<i_max;i++){
                 for(j=i_0;j<j_max;j++){
                     m=se+sed*RANSU;
                     nl=D[i][j]*(1-D[i][j])*((D[i][j]-0.5)*ga+m);
                     D2[i][j]=D[i][j]+tee*nl*V(n[i][j]);
                 }
             }

//----- D & n diffusion term -----

             for(i=i_0;i<i_max;i++){
                 for(j=i_0;j<j_max;j++){
                     dD=D[i-1][j]+D[i+1][j]+D[i][j-1]+D[i][j+1]
                         -4.0*D[i][j];
                     D3[i][j]=D2[i][j]+txxg*dD*V(n[i][j]);

                     dn=n[i-1][j]+n[i+1][j]+n[i][j-1]+n[i][j+1]
                         -4.0*n[i][j];

                     nn[i][j]=n[i][j]+txx*dn;
                 }
             }

//----- n local map -----
             for(i=i_0;i<i_max;i++){
                 for(j=i_0;j<j_max;j++){
                     n[i][j]=nn[i][j]-DT*D[i][j]*n[i][j];
                     if( n[i][j] < 0 ) n[i][j]=0.0;
                  
                     D[i][j]=D3[i][j];
                 }
             }

             boundary_condition();
         }

         t = t + time_interval;

         if(*(argv[1]+1)=='y'){

             step=integer(t/DT);
             if((step%save_step)==0){
                 n_snapshot_save((int)(step/save_step));
                 if(*argv[1]!='B') cout<<"snapshot_save t= "<<t<<"\n";
             }
             if(step >= quit_step) break;
         }

         if(*argv[1]!='B') win_draw();

      }
    }

}


//************ boundary condition ************

void boundary_condition(void)
{
    register int i,j;

/*  periodic */
/*
    for(j=i_0;j<j_max;j++){
        n[i_0-1][j]=n[i_max-1][j];
        n[i_max][j]=n[i_0][j];
    }
    for(i=i_0;i<i_max;i++){
        n[i][i_0-1]=n[i][j_max-1];
        n[i][j_max]=n[i][i_0];
    }
*/

/* free + fix */

    for(j=i_0;j<j_max;j++){
        n[i_0-1][j]=n[i_0][j];
        n[i_max][j]=1.0;
        D[i_0-1][j]=D[i_0][j];
        D[i_max][j]=0;

    }
    for(i=i_0;i<i_max;i++){
        n[i][i_0-1]=n[i][i_0];
        n[i][j_max]=1.0;
        D[i][i_0-1]=D[i][i_0];
        D[i][j_max]=0;
    }


/* trous */
/*  for(j=i_0;j<j_max;j++){
        n[i_0-1][j]=n[i_0][j];
        n[i_max][j]=0.0;
    }
    for(i=i_0;i<i_max;i++){
        n[i][i_0-1]=n[i][j_max-1];
        n[i][j_max]=n[i][i_0];
    }
*/

}


//*************** initialize function ************

void initialize_function(void)
{
    register int i,j;
    double xi;

    for(j=i_0; j<j_max; j++){
        for(i=i_0; i<i_max; i++){
            D[i][j] = 0;
            n[i][j] = 1.0;
        }
    }

    for(j=i_0; j<j_max; j++)
        for(i=i_0; i<i_max; i++)
            if(((i-i_0)*(i-i_0)+(j-i_0)*(j-i_0))<(r_start/DX*r_start/DX))
                D[i][j] = 1;

    boundary_condition();

}


//************** window draw **************

void win_draw(void)
{
    register int i,j;

    if (changedisp==0){

        for(j=i_0; j<j_max; j++)
            for(i=i_0; i<i_max; i++)
                X.color_draw(D[i][j],i,j);
            
    }else if (changedisp==1){

        for(j=i_0; j<j_max; j++)
             for(i=i_0; i<i_max; i++)
                 X.color_draw(n[i][j],i,j);

    }else{

    }

    return;
}


//************ number n snapshot_save ************

void n_snapshot_save(int n)
{
    char number[4];
    char name0[26];
    char *filename;

    sprintf(name0,"%s",name);
    sprintf(number,"%d",n);
    filename=strcat(name0,number);
    snapshot_save(filename);
    return;
}


//*********** snapshot_save **************

void  snapshot_save(char *filename)
{
      static int use=0;          /* stop at too many saves */
      ++use;
      if( use > 100 ) exit(1);

      ofstream file(filename);
      
      file<<"t= "<<t<<"\n";
      for(int j=i_0; j<j_max; j++)
          for(int i=i_0; i<i_max; i++)
              file<<i<<" "<<j<<" "<<D[i][j]<<" "<<n[i][j]<<"\n";
      
      file.close();
}


//*********** snapshot_load *************

void  snapshot_load(char *filename)
{
      char s[126];
      int k;
      
      ifstream file(filename);
      if(!file){
          cout<<" can't open "<<filename<<"\n";
          exit(1);
      }

      file>>s>>t;
      for(int j=i_0; j<j_max; j++)
          for(int i=i_0; i<i_max; i++)
              file>>k>>k>>D[i][j]>>n[i][j];
      
      file.close();
}


//************ read_parameter **************

void read_parameter(char *filename)
{
    char s[126];

    ifstream file(filename);
    if(!file){
       cout<<" can't open "<<filename<<"\n";
       exit(1);
    }

    file>>s;
    file>>s>>LX;
    file>>s>>LY;
    file>>s>>DT;
    file>>s>>DX;
    file>>s>>disturb;
    file>>s>>alpha;
    file>>s>>beta;
    file>>s>>e;
    file>>s>>ga;
    file>>s>>save_time;
    file>>s>>quit_time;
    file>>s>>name;
    file>>s>>r_start;
    file>>s>>changedisp;

    file.close();
    return;
}


          


    